module dataio

    interface
        subroutine readengfc( uid, engf, mapf, tqf, ftqf, nidle, nrate, nex )
            integer uid
            real(8) nidle, nrate, nex
            character (len=*) engf, mapf, tqf, ftqf
        end subroutine

        subroutine readgear( uid, tmf, ngr, gr, nsgr, sgr, grs, swat )
            integer uid, ngr, nsgr, grs, swat
            character (len=*) tmf
            real(8), pointer :: gr(:), sgr(:)
        end subroutine

        subroutine readspecfc ( uid, specname, rt, rt2, fgr, tpname, vhname, engf, tmf, tsout, rrc, cd,&
            istop, t_istop, istop2, t_istop2, t_istop_off, Kf1, Kf2)
            integer uid, tsout, istop, t_istop, istop2, t_istop2, t_istop_off
            real(8) rt, rt2, fgr, rrc, cd, Kf1, Kf2
            character (len=*) specname, tpname, vhname, engf, tmf
        end subroutine

        subroutine readtqfc ( uid, tqf, tqc )
            use hvtransfc
            integer uid
            type (stqcurve) tqc
            character (len=*) tqf
        end subroutine readtqfc

        subroutine readmapfc( uid, mapf, fcm )
            use hvtransfc
            integer uid
            type (sfcmap) fcm
            character (len=*) mapf
        end subroutine readmapfc

        subroutine outaveragefc ( uid, fname, specf, vhname, tpname, engf, tmf,     &
                                  fgr, rt, rt2, avvu, eu, avvh, eh, avvt, et, ric, ece, Kf1, Kf2, &
                                  istop, istop2, t_istop, t_istop2, t_istop_off )
            integer uid, istop, istop2, t_istop, t_istop2, t_istop_off
            character (len=*) fname, specf, vhname, tpname, engf, tmf
            real(8) fgr, rt, rt2, eu, avvu, eh, avvh, et, avvt, ric, ece, Kf1, Kf2
        end subroutine

        subroutine outtimeseriesfc (uid,fname,vpat,vreal,ne,te,nne,nte,sp,fc,comments)
            use hvtransfc
            integer uid, sp(:)
            character (len=*) fname, comments
            type (scycle) vpat
            real(8) vreal(:), ne(:), te(:), fc(:), nne(:), nte(:)
        end subroutine
        
        subroutine check_results (vpat,vreal,ne,te,nne,nte,sp,fc,comments)
            use hvtransfc
            implicit none
            integer uid, i, sp(:)
            character (len=*) comments
            type (scycle) vpat
            real(8) vreal(:), ne(:), te(:), fc(:), nne(:), nte(:)
        end subroutine check_results
        

    end interface

end module dataio




!**************************************************************
! SUBROUTINE READENGFC
!  Read engine input data (map, torque, engine speed)
!**************************************************************
subroutine readengfc( uid, engf, mapf, tqf, ftqf, nidle, nrate, nex )
    implicit none

    integer uid
    character (len=*) tqf, ftqf, mapf, engf
    real(8) nidle, nrate, nex

    open ( uid, file = engf, status = 'old', access = 'sequential', err=880 )

    read ( uid, *, err=890 ) mapf                               ! fuel map
    read ( uid, *, err=890 ) tqf                                ! max torque
    read ( uid, *, err=890 ) ftqf                               ! frictional torque
    read ( uid, *, err=890 ) nidle                              ! idling speed
    read ( uid, *, err=890 ) nrate                              ! rated speed
    read ( uid, *, err=890 ) nex                                ! max speed

    close ( uid )

    return

!   -----------------------------------------------------------
880 write (0,'(3A)') " Cannot open engine file [ ", trim(engf), " ]"
    stop
890 write (0,'(3A)') " Failed to read engine file [ ", trim(engf), " ]"
    close ( uid )
    stop

end subroutine




!**************************************************************
! SUBROUTINE READGEAR
!  Read transmission input data
!**************************************************************
subroutine readgear( uid, tmf, ngr, gr, nsgr, sgr, grs, swat )
    implicit none

    integer uid, i, ngr, nsgr, grs, swat
    character(len=*) tmf                                        ! transmission filename
    real(8), pointer :: gr(:), sgr(:)

    open ( uid, file = tmf, status = 'old', access = 'sequential', err=900 )

    read ( uid, *, err=910 ) grs                                ! start gear
    read ( uid, *, err=910 ) ngr                                ! n of main gear

    allocate ( gr(ngr) )
    do i = 1, ngr
        read ( uid, *, err=910 ) gr(i)                          ! gear ratio
    end do

    read ( uid, *, err=910 ) nsgr                               ! n of sub transmission

    allocate( sgr( nsgr ) )
    do i = 1, nsgr
        read ( uid, *, err=910 ) sgr(i)                         ! gear ratio
    end do

    read ( uid, *, err=910 ) swat                               ! torque converter AT(0:MT,AMT, 1:AT)

    close( uid )
    return

!   -----------------------------------------------------------
900 write (0,'(3A)') " Cannot open transmission file [ ", trim(tmf), " ]"
    stop
910 write (0,'(3A)') " Failed to read transmission file [ ", trim(tmf), " ]"
    close( uid )
    stop

end subroutine



!**************************************************************
! SUBROUTINE READSPECFC
!  Read vehicle specification data
!**************************************************************
subroutine readspecfc ( uid, specf, rt, rt2, fgr, tpname, vhname, engf, tmf, tsout, rrc, cd,&
    istop, t_istop, istop2, t_istop2, t_istop_off, Kf1, Kf2 )
    implicit none

    character (len=*) specf, tpname, vhname, engf, tmf
    integer uid, ios, tsout, istop, t_istop, istop2, t_istop2, t_istop_off
    real(8) rt, rt2, fgr, rrc, cd, Kf1, Kf2

    open ( uid, file=trim(specf), status='OLD', access='sequential', iostat=ios )
    if ( ios /= 0 ) then
        write(0,*) ' Cannot open spec-file : ', trim(specf)
        stop
    end if

    read ( uid, *, err=650 ) vhname                                      ! vehicle name
    read ( uid, *, err=650 ) tpname                                      ! type name
    read ( uid, *, err=650 ) engf                                        ! engine filename
    read ( uid, *, err=650 ) tmf                                         ! transmission filename
    read ( uid, *, err=650 ) fgr                                         ! final ratio
    read ( uid, *, err=650 ) rt                                          ! tire radius (m)
    read ( uid, *, err=650 ) rt2                                         ! ő^Ca(m)
    read ( uid, *, err=650 ) tsout                                       ! time series output 0:OFF, 1:ON
    read ( uid, *, err=650 ) rrc                                         ! tire rolling resistance coefficient
    read ( uid, *, err=650 ) cd                                          ! air drag coefficient
    read ( uid, *, err=650 ) istop                                       ! flag_idling stop
    read ( uid, *, err=650 ) t_istop
    read ( uid, *, err=650 ) istop2                                      ! flag_idling stop(JE05X^[gj
    read ( uid, *, err=650 ) t_istop2
    read ( uid, *, err=650 ) t_istop_off                                 ! ԗiENGn^C~Oi5sj
    read ( uid, *, err=650 ) Kf1                                         ! Kf for urban
    read ( uid, *, err=650 ) Kf2                                         ! Kf for highway
    close(uid)
    
    if(istop == 0) t_istop = 999                                         ! No idlling_stop
    if(istop2 == 0) t_istop2 = 999
!    t_istop_off = 5                                    
    rrc = rrc * dsqrt(1.0_8/(1.0_8 + rt2))                               ! RH␳
    
    return

!   -----------------------------------------------------------
!   G[Ή
!   -----------------------------------------------------------
650 write (0,'(A,i0,3A)') " Failed to read spec_data file "
    close ( uid )
    stop

end subroutine




!**************************************************************
! SUBROUTINE READTQFC
!  Read engine torque input data
!**************************************************************
subroutine readtqfc ( uid, tqf, tqc )
    use hvtransfc
    implicit none

    integer uid
    character (len=*) tqf                                       ! torque filename
    type (stqcurve) tqc                                         ! torque curve

    character (len=1024) tmp
    integer i, j, gap, ios
    real(8) revt, tmaxt

!   -----------------------------------------------------------
!   count number of torque data
!   -----------------------------------------------------------
    tqc%ndata = 0                                               ! initialize
    open ( uid, file = tqf, status = 'old', err=600 )
    read ( uid,'(A)' ) tmp                                      ! skip header
    read ( uid,'(A)' ) tmp                                      ! skip header

    ios = 0                                                     ! status
    do while ( ios == 0 )
        read ( uid, *, iostat = ios, err=650 ) revt, tmaxt      ! test reading of data set
        if( ios == 0 ) tqc%ndata = tqc%ndata + 1                ! count n of data
    end do

!   -----------------------------------------------------------
!   read torque data
!   -----------------------------------------------------------
    allocate ( tqc%tq ( tqc%ndata ) )
    allocate ( tqc%rev( tqc%ndata ) )

    rewind ( uid )                                              ! move to head of file
    read ( uid,'(A)' ) tmp                                      ! skip header
    read ( uid,'(A)' ) tmp                                      ! skip header

    do i = 1, tqc%ndata
        read ( uid, *, err=650 ) tqc%rev(i), tqc%tq(i)          ! read torque data
    end do

    close ( uid )

!   -----------------------------------------------------------
!   sort in ascending order
!   -----------------------------------------------------------
    gap = ( tqc%ndata + 1) / 2
    do while ( gap /= 0 )
        i = gap
        do while ( i <= tqc%ndata )
            j = i - gap
            if (j <= 0) goto 100
            do while ( ( j >= 1 ) .and. ( tqc%rev( j ) > tqc%rev( j + gap ) ) )
                revt               = tqc%rev(j)
                tqc%rev( j )       = tqc%rev(j+gap)
                tqc%rev( j + gap ) = revt
                tmaxt              = tqc%tq( j )
                tqc%tq( j )        = tqc%tq( j + gap )
                tqc%tq( j + gap )  = tmaxt
                j = j - gap
            if (j <= 0) goto 100
            end do
100         i = i + 1
        end do
        gap = gap / 2
    end do

    return

!   -----------------------------------------------------------
!   G[Ή
!   -----------------------------------------------------------
600 write (0,'(3A)') " Cannot open torque data file [ ", trim(tqf), " ]"
    stop
650 write (0,'(A,i0,3A)') " Failed to read torque data. Data No.= ",    &
                          tqc%ndata+1, " in [ ", trim( tqf ), " ]"
    close ( uid )
    stop

end subroutine readtqfc

!**************************************************************
! SUBROUTINE READShiftPos
!  Read engine torque input data
!**************************************************************
subroutine readshiftpos ( uid, spf, tqc )
    use hvtransfc
    implicit none

    integer uid
    character (len=*) spf                                       ! Shift filename
    type (stqcurve) tqc                                         ! torque curve

    character (len=1024) tmp
    integer i, j, gap, ios
    real(8) revt, tmaxt

!   -----------------------------------------------------------
!   count number of torque data
!   -----------------------------------------------------------
    tqc%ndata = 0                                               ! initialize
    open ( uid, file = spf, status = 'old', err=600 )
    read ( uid,'(A)' ) tmp                                      ! skip header
    read ( uid,'(A)' ) tmp                                      ! skip header

    ios = 0                                                     ! status
    do while ( ios == 0 )
        read ( uid, *, iostat = ios, err=650 ) revt, tmaxt      ! test reading of data set
        if( ios == 0 ) tqc%ndata = tqc%ndata + 1                ! count n of data
    end do

!   -----------------------------------------------------------
!   read torque data
!   -----------------------------------------------------------
    allocate ( tqc%tq ( tqc%ndata ) )
    allocate ( tqc%rev( tqc%ndata ) )

    rewind ( uid )                                              ! move to head of file
    read ( uid,'(A)' ) tmp                                      ! skip header
    read ( uid,'(A)' ) tmp                                      ! skip header

    do i = 1, tqc%ndata
        read ( uid, *, err=650 ) tqc%rev(i), tqc%tq(i)          ! read torque data
    end do

    close ( uid )

!   -----------------------------------------------------------
!   sort in ascending order
!   -----------------------------------------------------------
    gap = ( tqc%ndata + 1) / 2
    do while ( gap /= 0 )
        i = gap
        do while ( i <= tqc%ndata )
            j = i - gap
            if (j <= 0) goto 100
            do while ( ( j >= 1 ) .and. ( tqc%rev( j ) > tqc%rev( j + gap ) ) )
                revt               = tqc%rev(j)
                tqc%rev( j )       = tqc%rev(j+gap)
                tqc%rev( j + gap ) = revt
                tmaxt              = tqc%tq( j )
                tqc%tq( j )        = tqc%tq( j + gap )
                tqc%tq( j + gap )  = tmaxt
                j = j - gap
            if (j <= 0) goto 100
            end do
100         i = i + 1
        end do
        gap = gap / 2
    end do

    return

!   -----------------------------------------------------------
!   G[Ή
!   -----------------------------------------------------------
600 write (0,'(3A)') " Cannot open torque data file [ ", trim(spf), " ]"
    stop
650 write (0,'(A,i0,3A)') " Failed to read torque data. Data No.= ",    &
                          tqc%ndata+1, " in [ ", trim( spf ), " ]"
    close ( uid )
    stop

end subroutine readshiftpos


!**************************************************************
! SUBROUTINE READMAPFC
!  R}bvf[^̓ǂݍ݃Tu[`
!   uid     : uԍ
!   mapf    : fuel map filename
!   fcm     : R}bv(\)
!**************************************************************
subroutine readmapfc( uid, mapf, fcm )
    use hvtransfc
    implicit none

    integer uid
    character (len=*) mapf                                      ! fuel map filename
    type (sfcmap) fcm                                           ! fuel map

    character (len=1024) tmp
    real(8) idles, idlet, r, t, f
    integer gap, i, j, k, ios, head, ct

!   -----------------------------------------------------------
!   count number of fuel map data
!   -----------------------------------------------------------
    fcm%ndata = 0                                               ! initialize n of data
    open( uid, file = mapf, status = 'old', err=500 )

    read ( uid, *, err=550 ) tmp                                ! skip header
    read ( uid, *, err=550 ) tmp                                ! skip header
    read ( uid, *, err=550 ) idles, idlet, fcm%fcidle           ! idling fuel consumption(L/h)

    ios = 0
    do while ( ios == 0 )
        read( uid, *, iostat = ios, err=550 ) r, t, f           ! test reading
        if( ios == 0 ) fcm%ndata = fcm%ndata + 1                ! count n of data
    end do

!   -----------------------------------------------------------
!   read fuel map
!   -----------------------------------------------------------
    allocate( fcm%rev( fcm%ndata ) )
    allocate( fcm%tq( fcm%ndata )  )
    allocate( fcm%fc( fcm%ndata )  )

    rewind( uid )                                               ! move to head of file
    read ( uid, *, err=550 ) tmp                                ! skip header
    read ( uid, *, err=550 ) tmp                                ! skip header
    read ( uid, *, err=550 ) tmp                                ! skip idling fuel

    do i = 1, fcm%ndata                                         ! read fuel map
        read ( uid, *, err=550 ) fcm%rev(i), fcm%tq(i), fcm%fc(i)
    end do

    close( uid )

! ------ SORT MAP DATA WITH ENGINE SPEED -----------------------
    gap = ( fcm%ndata + 1 ) / 2
    do while ( gap > 0 )
        i = gap
        do while ( i <= fcm%ndata )
            j = i - gap
            if (j <= 0) goto 100
            do while( ( j >= 1 ) .and. ( fcm%rev(j) > fcm%rev(j+gap) ) )
                r              = fcm%rev(j)
                fcm%rev(j)     = fcm%rev(j+gap)
                fcm%rev(j+gap) = r
                t              = fcm%tq(j)
                fcm%tq(j)      = fcm%tq(j+gap)
                fcm%tq(j+gap)  = t
                f              = fcm%fc(j)
                fcm%fc(j)      = fcm%fc(j+gap)
                fcm%fc(j+gap)  = f
                j = j - gap
            if (j <= 0) goto 100
            end do
100         i = i + 1
        end do
        gap = gap / 2
    end do

! ------ SORT WITH ENGINE TORQUE -------------------------------
    head = 1
    r = fcm%rev( head )
    ct = 0
    do i = 1, fcm%ndata
        ct = ct + 1
        if ( ( r /= fcm%rev(i) ) .or. ( i == fcm%ndata ) ) then
            if ( r /= fcm%rev(i) ) then
                ct = ct -1
            end if
            gap = ( ct + 1 ) / 2
            do while ( gap > 0 )
                k = gap
                do while ( k <= ct - 1 )
                    j = head + k - gap
                    if (j < head) goto 110
                    do while ( ( j >= head ) .and. ( fcm%tq(j) > fcm%tq(j+gap) ) )
                        r              = fcm%rev(j)
                        fcm%rev(j)     = fcm%rev(j+gap)
                        fcm%rev(j+gap) = r
                        t              = fcm%tq(j)
                        fcm%tq(j)      = fcm%tq(j+gap)
                        fcm%tq(j+gap)  = t
                        f              = fcm%fc(j)
                        fcm%fc(j)      = fcm%fc(j+gap)
                        fcm%fc(j+gap)  = f
                        j = j - gap
                    if (j < head) goto 110
                    end do
110                 k = k + 1
                end do
                gap = gap / 2
            end do
            r = fcm%rev(i)
            ct = 1
            head=i
        end if
    end do

    return

!   -----------------------------------------------------------
!   error handling
!   -----------------------------------------------------------
500 write (0,'(3A)') " Cannot open map data file [ ", trim(mapf), " ]"
    stop
550 write (0,'(3A)') " Failed to read fuel map [ ", trim(mapf), " ]"
    close ( uid )
    stop

end subroutine




!**************************************************************
!  SUBROUTINE OUTAVERAGEFC
!   Output result of average speed & fuel consumption
!**************************************************************
subroutine outaveragefc ( uid, outf, specf, vhname, tpname, engf, tmf, &
                          fgr, rt, rt2, avvu, eu, avvh, eh, avvt, et, ric, ece, Kf1, Kf2, &
                          istop, istop2, t_istop, t_istop2, t_istop_off )
    use hvtransfc
    implicit none

    character, parameter :: ht = char(9)

    integer uid, istop, istop2, t_istop, t_istop2, t_istop_off
    character (len=*) outf, specf, vhname, tpname, engf, tmf
    real(8) fgr, rt, rt2, eu, avvu, eh, avvh, avvt, et, ric, ece, Kf1, Kf2
    real(8) eu_kato, eu_Kf, et_kato, et_Kf, eh_Kf
    real(8) ece_kf, ece_kato

! ߓn␳
    eu_kato = eu / 1.03_8  ! urban transient correction 3%
    et_kato = et / 1.03_8  ! mid-town transient correction 3%
    eu_Kf = eu_kato * Kf1  ! ߓn␳͏Ŏ{ς
    et_Kf = et_kato * Kf1
    eh_Kf = eh * Kf2
    ece_kf = 1.0_8 / ((1.0_8 - ric) / eu_Kf + ric / eh_Kf)  

    open ( uid, file = outf, status = 'unknown' )
    
    write ( uid, '(3A)' ) "VEHICLENAME", ht, trim(vhname)
    write ( uid, '(3A)' ) "TYPE", ht, trim(tpname)
    write ( uid, '(3A)' ) "SPEC FILE", ht, trim(specf)
    write ( uid, '(3A)' ) "ENGINE FILE", ht, trim(engf)
    write ( uid, '(3A)' ) "TRANSMISSION FILE", ht, trim(tmf)
    write ( uid, '(2A,F8.5)' ) "FINAL GEAR RATIO", ht, fgr
    write ( uid, '(2A,F8.5)' ) "TIRE RADIUS(m)",    ht, rt
    write ( uid, '(2A,F8.5)' ) "RH␳p\^Caimj", ht, rt2
    write ( uid, '(2A,I2)' ) "AChXgbvi0:ȂC1:j",ht, istop
    write ( uid, '(2A,I2)'  ) "GW~܂ł̑҂(s)  ", ht, t_istop
    write ( uid, '(2A,I2)' ) "JE05[hJñAChXgbvi0:ȂC1:j", ht, istop2
    write ( uid, '(2A,I2)' ) "JE05[hJñAChXgbv쓮(s)", ht, t_istop2
    write ( uid, '(2A,I2)' ) "ԗĩGWn^C~O(s)", ht, t_istop_off
    write ( uid, * )
 
    write( uid, * ) "yURBANz"
    write ( uid, '(4A,F8.4,3A,F5.1)' ) "@ߓn␳O    ",ht,"FC(km/l)",ht,eu,ht,"Ave.Speed(km/h)",ht,avvu
    write ( uid, '(4A,F8.4,3A,F5.1)' ) "@ߓn␳    ",ht,"FC(km/l)",ht,eu_kato,ht,"3%␳ "
    write ( uid, '(4A,F8.4,3A,F8.6)' ) "@Kfll      ",ht,"FC(km/l)",ht,eu_Kf,ht,"Kf1      ",ht,Kf1
    
    write( uid,*)
        
    write( uid, * ) "yHIGHWAYz"
    write ( uid, '(4A,F8.4,3A,F5.1)' ) "       @       ",ht,"FC(km/l)",ht,eh,ht,"Ave.Speed(km/h)",ht,avvh
    write ( uid, '(4A,F8.4,3A,F8.6)' ) "@Kfll      ",ht,"FC(km/l)",ht,eh_Kf,ht,"Kf2      ",ht,Kf2
    
    write( uid,*)
        
    write( uid, * ) "yAVERAGEz"
    write ( uid, '(4A,F8.4,3A,F5.2)' ) "       @       ",ht,"FC(km/l)",ht,ece_Kf,ht,"HIGHWAY RATIO",ht, ric
    
    write( uid,*)
        
    write( uid, * ) "yMID-TOWNz"
    write ( uid, '(4A,F8.4,3A,F5.1)' ) "@ߓn␳O    ",ht,"FC(km/l)",ht,et,ht,"Ave.Speed(km/h)",ht,avvt
    write ( uid, '(4A,F8.4,3A,F5.1)' ) "@ߓn␳    ",ht,"FC(km/l)",ht,et_kato,ht,"3%␳ "
    write ( uid, '(4A,F8.4,3A,F8.6)' ) "@Kfll      ",ht,"FC(km/l)",ht,et_Kf,ht,"Kf1      ",ht,Kf1
 
 
    close ( uid )

    end subroutine outaveragefc

!**************************************************************
!  SUBROUTINE OUTPUTTIMESERIES
!   Output fuel consumption data of time series
!**************************************************************
subroutine outtimeseriesfc (uid,outf,vpat,vreal,ne,te,nne,nte,sp,fc,comments)
    use hvtransfc
    implicit none

    character, parameter :: ht = char(9)

    integer uid, i, sp(:)
    real(8) vreal(:), ne(:), te(:), fc(:), nne(:), nte(:)
    character (len=*) outf, comments
    type (scycle) vpat

    open ( uid, file = outf, status = 'unknown', position = 'APPEND' )

    write ( uid, * )
    write ( uid, '(A)' ) trim(comments)
    write ( uid, '(17A)' ) 'time(s)', ht, 'Vtarget(km/h)', ht, 'Vreal(km/h)', &
            ht, 'Ne(rpm)', ht, 'Te(N-m)', ht, 'N_norm(%)', ht, 'T_norm(%)',   &
            ht, 'Shift', ht, 'FC(l/h)'

    do i = 1, vpat%ndat
        write (uid, '(i0,2(a,f0.4),2(a,f0.1),2(a,f0.2),a,i0,a,f0.6)')   &     ! f0.2 -> f0.4@o͌
               i, ht, vpat%v(i), ht, vreal(i), ht, ne(i), ht, te(i),    &
               ht, nne(i), ht, nte(i), ht, sp(i), ht, fc(i)
    end do

    close ( uid )

end subroutine outtimeseriesfc

!**************************************************************
!  SUBROUTINE check_results
!  Check output data of time series
!**************************************************************
subroutine check_results (vpat,vreal,ne,te,nne,nte,sp,fc,comments)
    use hvtransfc
    implicit none

    character, parameter :: ht = char(9)

    integer uid, i, sp(:)
    real(8) vreal(:), ne(:), te(:), fc(:), nne(:), nte(:)
    character (len=*) comments    !outf 
    type (scycle) vpat

    do i = 1, vpat%ndat
        if ((vreal(i)-vpat%v(i))>0.01) then
            write(*,*) 'error: V_real > V_target'
!            stop
        end if
        if (nte(i)>100.)then
            write(*,*) 'error: te > 100% ', i
!            stop
        end if        
    end do
end subroutine check_results