module vehicleparams

    interface
        subroutine setparameters( tpname, spec, eng, tm, gr, sgr, grs, rrc, cd )
            use hvtransfc
            character (len=*) tpname
            type (stransmission) tm
            type (sengine) eng
            type (sspec) spec
            real(8), pointer :: gr(:), sgr(:)
            integer grs
            real(8) rrc, cd
        end subroutine

        subroutine showinputdata( spec, eng, tm, vhname, tpname )
            use hvtransfc
            type (sspec) spec
            type (sengine) eng
            type (stransmission) tm
            character (len=*) vhname, tpname
        end subroutine showinputdata
    end interface

end module vehicleparams



!*********************************************************************
! READ SPEC DATA, VEHICLE, ENGINE, TRANSMISSION, MAX&FRICTION TORQUE
!*********************************************************************
subroutine setparameters( tpname, spec, eng, tm, gr, sgr, grs, rrc, cd )
    use hvtransfc
    use modconst
    implicit none

    character (len=*) tpname
    type (stransmission) tm                                     ! transmission
    type (sengine) eng                                          ! engine
    type (sspec) spec                                           ! vehicle spec
    real(8), pointer :: gr(:), sgr(:)                           ! gear ratio
    integer i, j, x, grs
    real(8) dwt, dwe                                            ! rotational inertia
    real(8) nerange                                             ! nrate - nidle
    real(8) rrc, cd         ! ]RW@@CRW

    real(8) beta ! ]ʂ̕␳W
    
!   -----------------------------------------------------------
!   spec
!   -----------------------------------------------------------
    call setvehicletype( tpname, spec )

    spec%tcl = spec%rt * 2.0_8 * 3.14_8                         ! tire circumference(m)
    spec%gvw = spec%w0 + spec%wld + REAL(spec%crew, 8) * PSGW   ! GVW (kg)

    select case ( tpname(1:1) )
    case ( 'T' )
        spec%wt = spec%w0 + spec%wld * spec%rld + 55.0_8
!        spec%mur = ( 0.00513_8 + 17.6_8 / spec%wt ) * 9.8_8
!        spec%mua = ( 0.00299 * spec%bh * spec%bw - 0.000832_8 ) * 9.8_8
        spec%mur = rrc * 9.8_8 + (0.00023 + 6.7/spec%wt)*9.8_8   ! 쓮nR0.00023+EEE
        spec%mua = cd * spec%bh * spec%bw   ! Cd [N/m2/(km/h)2]
        
    case ( 'B' )
        spec%wt = spec%w0 + spec%crew * 55.0_8 * spec%rld
!        spec%mur = ( 0.00513_8 + 17.6_8 / spec%wt ) * 9.8_8
!        spec%mua = ( 0.00299 * spec%bh * spec%bw - 0.000832_8 ) * 0.680_8 * 9.8_8
         spec%mur = rrc * 9.8_8 + (0.00023 + 6.7/spec%wt)*9.8_8   ! 쓮nR0.00023+EEE
         spec%mua = cd * spec%bh * spec%bw   ! Cd [N/m2/(km/h)2]
    case default
        write (0, '(3A)')                                       &
           " Vehicle Type Error, [ ", trim(tpname), " ] is not defined."
        stop
    end select

!   -----------------------------------------------------------
!   engine
!   -----------------------------------------------------------
    nerange = eng%nrate - eng%nidle                             ! nrate - nidle
    eng%nes = PMEET * nerange + eng%nidle                       ! clutch meet, rpm
    eng%nec = PRELEASE * nerange + eng%nidle                    ! clucth off, rpm

!   -----------------------------------------------------------
!   transmission
!   -----------------------------------------------------------
!    tm%grs = grs * tm%nsgr                                     ! start gear
    tm%grs = grs                                                ! start gear
    if( tm%at == 1) then
        tm%grb = 1
    else
        tm%grb = ( GRB - 1 ) * tm%nsgr + 1                      ! bottom gear
    end if
    tm%skiplimit = ( skiplimit - 1 ) * tm%nsgr                  ! limit of skip shift
    tm%efgr = EGR_ND                                            ! efficiency of final
    tm%ngr = tm%nmgr * tm%nsgr                                  ! total number of shift position

!   -----------------------------------------------------------
!   parameter setting for each gear
!   -----------------------------------------------------------
    allocate ( tm%gri( tm%ngr ) )                               ! for gear ratio
    allocate ( eng%nel( tm%ngr ) )                              ! for minimum speed
    dwt = spec%w0 * PDWT                                        ! rotational inertia of tire
    dwe = spec%w0 * PDWE                                        ! rotational inertia of engine

    x = 1                                                       ! parameter setting for each gear
    do i = 1, tm%nmgr
        do j = 1, tm%nsgr

            tm%gri(x)%gr = gr(i) * sgr(j)                       ! gear ratio
            tm%gri(x)%dw = dwt + spec%rite * (tm%gri(x)%gr)**2 &
                         * (tm%fgr)**2 / (spec%rt)**2           ! rotational mass

            if ( gr(i) == 1.0_8 ) then                          ! transmission efficiency     170606
!            if ( tm%gri(x)%gr == 1.0_8 ) then                   ! transmission efficiency   _mod. 170118
                tm%gri(x)%egr = EGR_DIRECT
            else
                tm%gri(x)%egr = EGR_ND
            end if

            if ( spec%gvw >= 8000.0_8 ) then                    ! torque margin
                if ( i >= grs+2 ) then                          ! GVW >= 8t
                    tm%gri(x)%tmargin = tmup8t(3)
                else if ( i <= grs ) then
                    tm%gri(x)%tmargin = tmup8t(1)
                else
                    tm%gri(x)%tmargin = tmup8t(2)
                end if
            else                                                ! torque margin
                if ( i >= grs+2 ) then                          ! GVW >= 8t
                    tm%gri(x)%tmargin = tmunder8t(3)
                else if ( i <= grs ) then
                    tm%gri(x)%tmargin = tmunder8t(1)
                else
                    tm%gri(x)%tmargin = tmunder8t(2)
                end if
            end if

            if ( spec%gvw >= 8000.0_8 ) then                           ! minimum speed
                if ( i >= grs+3 ) then                                 ! GVW >= 8t
                    eng%nel(x) = nminsup8t(4) * nerange + eng%nidle
                else if ( i <= grs ) then
                    eng%nel(x) = nminsup8t(1) * nerange + eng%nidle
                else
                    eng%nel(x) = nminsup8t(i-grs+1) * nerange + eng%nidle
                end if
            else                                                       ! minimum speed
                if ( i >= grs+3 ) then                                 ! GVW < 8t
                    eng%nel(x) = nminsun8t(4) * nerange + eng%nidle
                else if ( i <= grs ) then
                    eng%nel(x) = nminsun8t(1) * nerange + eng%nidle
                else
                    eng%nel(x) = nminsun8t(i-grs+1) * nerange + eng%nidle
                end if
            end if

            x = x + 1
        end do
    end do


end subroutine



!**************************************************************
!  SHOWINPUTDATA
!   Display vehicle spec & input parameters
!**************************************************************
subroutine showinputdata( spec, eng, tm, vhname, tpname )
    use hvtransfc
    implicit none

    type (sspec) spec
    type (sengine) eng
    type (stransmission) tm
    character (len=*) vhname, tpname
    integer i

    print*, "----------------------------------------------------------"
    print '(2A)', " Vehicle Name    : ", trim(vhname)
    select case ( tpname(1:1) )
    case ( 'T' )
        print '(A)', " Vehicle type  : TRUCK"
    case ( 'B' )
        print '(A)', " Vehicle type  : BUS"
    end select
    print '(2A)',    " Category      : ", trim(tpname)
    if( tm%at == 1 ) then
        print '(A)', " Transmission  : Torque Converter AT"
    else
        print '(A)', " Transmission  : MT, AMT"
    end if
    print '(A,F0.2,A)', " HIGHWAY RATIO : ", spec%ric * 100.0_8, " [%]"
    print*

    print*, "----------------------------------------------------------"
    print '(A,F8.2,A)', " GVW =", spec%gvw, "[kg]"
    print '(A,F8.2,A,F8.2,A)', " Wcurb =", spec%w0, "[kg], Wtest =", spec%wt, "[kg]"
    print '(A,F8.3,A,F8.3,A,F8.3,A)', " Width =", spec%bw,              &
          "[m],  Height=" , spec%bh, "[m], Tire radius=",spec%rt, "[m]"
    print '(A,I3)'," Crew =", spec%crew
    print*
    print '(A,F8.2,A,F8.2,A,F8.2,A)', " Nidle =", eng%nidle,            &
          "[rpm], Nrate =", eng%nrate, "[rpm], Nex  =", eng%nex, "[rpm]"
    print '(A,F8.2,A,F8.2,A)', " Nes   =", eng%nes,                     &
          "[rpm], Nec   =", eng%nec, "[rpm]"
    print '(A,F10.6,A)', " MuAir =", spec%mua, " [N/(km/h)^2]"
    print '(A,F10.6,A)', " MuRoll =", spec%mur, " [N/kg]"
    print*

    print '(A,I3)', " NUMBER OF GEAR =", tm%ngr
    print '(A)', " GEAR   RATIO  EFFICIENCY  TORQ MARGIN    DW[kg]"
    do i = 1, tm%ngr
        print '(I4,A,F8.3,F10.3,F12.3,F15.5)', i, ": ", tm%gri(i)%gr,   &
              tm%gri(i)%egr, tm%gri(i)%tmargin, tm%gri(i)%dw
    end do
    print '(A,F8.3,F10.3)', " FIN: ", tm%fgr, tm%efgr
    print*

end subroutine showinputdata



! *************************************************************
! SUBROUTINE TOUPPER
!  convert small letters to capital letters
!  buf : converted strings
! *************************************************************
subroutine toUpper( buf )
    implicit none
    integer strlength, i
    character ( len = * ) buf

    strlength = len( buf )
    do i = 1, strlength
        if( ichar(buf(i:i))>=97 .and. ichar(buf(i:i))<=122 ) then
            buf(i:i) = char(ichar(buf(i:i))-32)
        end if
    end do

end subroutine



! *************************************************************
! SUBROUTINE SETVEHICLETYPE
!  set parameters from standard vehicle category name
!  typename : vehicle category name
!  spec : vehicle spec
! *************************************************************
subroutine setvehicletype( categoryname, spec )
    use hvtransfc
    implicit none
    type (sspec) :: spec
    character (len=*) :: categoryname

    call toUpper( categoryname )                                ! convert category string to capital

    select case ( categoryname )
!   ----- TRUCK TRACTOR, "TTx" ---------------------------------
    case ("TT1")                                                ! TT1 GVW <=20t
        spec%w0 = 12300.0_8
        spec%wld = 29431.0_8
        spec%crew = 2
        spec%bh = 3.266_8
        spec%bw = 2.490_8
        spec%ric = 0.45_8
        spec%rld = 0.50_8       ! Load ratio
        spec%rite = 2.544_8     ! Ite

    case ("TT2")                                                ! TT2 GVW >20t
        spec%w0 = 19421.0_8
        spec%wld = 38910.0_8
        spec%crew = 2
        spec%bh = 3.191_8
        spec%bw = 2.490_8
        spec%ric = 0.45_8
        spec%rld = 0.50_8       ! Load ratio
        spec%rite = 2.544_8     ! Ite


!   ----- GENERAL TRUCK, "Tx" ----------------------------------
    case ("T1")                                                 ! T1 3.5t< GVW <= 7.5t
        spec%w0 = 2097.0_8                                      ! Payload <=1.5t
        spec%wld = 1482.0_8
        spec%crew = 3
        spec%bh = 1.991_8
        spec%bw = 1.717_8
        spec%ric = 0.15_8
        spec%rld = 0.45_8       ! Load ratio
        spec%rite = 0.270_8     ! Ite


    case ("T2")                                                 ! T2 3.5t< GVW <= 7.5t
        spec%w0 = 2496.0_8                                      ! 1.5t < Payload <=2t
        spec%wld = 2000.0_8
        spec%crew = 3
        spec%bh = 2.077_8
        spec%bw = 1.819_8
        spec%ric = 0.15_8
        spec%rld = 0.45_8       ! Load ratio
        spec%rite = 0.315_8     ! Ite


    case ("T3")                                                 ! T3 3.5t< GVW <= 7.5t
        spec%w0 = 2750.0_8                                      ! 2t < Payload <=3t
        spec%wld = 2999.0_8
        spec%crew = 3
        spec%bh = 2.153_8
        spec%bw = 1.989_8
        spec%ric = 0.15_8
        spec%rld = 0.45_8       ! Load ratio
        spec%rite = 0.315_8     ! Ite


    case ("T4")                                                 ! T4 3.5t< GVW <= 7.5t
        spec%w0 = 2913.0_8                                      ! 3t < Payload
        spec%wld = 3637.0_8
        spec%crew = 3
        spec%bh = 2.264_8
        spec%bw = 2.181_8
        spec%ric = 0.15_8
        spec%rld = 0.45_8       ! Load ratio
        spec%rite = 0.315_8     ! Ite


    case ("T5")                                                 ! T5 7.5t< GVW <= 8t
        spec%w0 = 3473.0_8
        spec%wld = 4239.0_8
        spec%crew = 2
        spec%bh = 2.471_8
        spec%bw = 2.303_8
        spec%ric = 0.35_8
        spec%rld = 0.50_8       ! Load ratio
        spec%rite = 0.703_8     ! Ite


    case ("T6")                                                 ! T6 8t< GVW <= 10t
        spec%w0 = 3663.0_8
        spec%wld = 6081.0_8
        spec%crew = 2
        spec%bh = 2.579_8
        spec%bw = 2.313_8
        spec%ric = 0.40_8
        spec%rld = 0.50_8       ! Load ratio
        spec%rite = 1.101_8     ! Ite


    case ("T7")                                                 ! T7 10t< GVW <= 12t
        spec%w0 = 4019.0_8
        spec%wld = 6380.0_8
        spec%crew = 2
        spec%bh = 2.536_8
        spec%bw = 2.343_8
        spec%ric = 0.40_8
        spec%rld = 0.50_8       ! Load ratio
        spec%rite = 1.101_8     ! Ite


    case ("T8")                                                 ! T8 12t< GVW <= 14t
        spec%w0 = 4788.0_8
        spec%wld = 8540.0_8
        spec%crew = 2
        spec%bh = 2.641_8
        spec%bw = 2.390_8
        spec%ric = 0.40_8
        spec%rld = 0.50_8       ! Load ratio
        spec%rite = 1.101_8     ! Ite


    case ("T9")                                                 ! T9 14t< GVW <= 16t
        spec%w0 = 5728.0_8
        spec%wld = 8684.0_8
        spec%crew = 2
        spec%bh = 2.672_8
        spec%bw = 2.391_8
        spec%ric = 0.40_8
        spec%rld = 0.50_8       ! Load ratio
        spec%rite = 1.101_8     ! Ite


    case ("T10")                                                ! T10 16t< GVW <= 20t
        spec%w0 = 8310.0_8
        spec%wld = 11109.0_8
        spec%crew = 2
        spec%bh = 3.043_8
        spec%bw = 2.490_8
        spec%ric = 0.40_8
        spec%rld = 0.50_8       ! Load ratio
        spec%rite = 1.650_8     ! Ite


    case ("T11")                                                ! T11 20t< GVW
        spec%w0 = 9193.0_8
        spec%wld = 14844.0_8
        spec%crew = 2
        spec%bh = 3.800_8       ! 3.8m Van
        spec%bw = 2.490_8
        spec%ric = 0.55_8
        spec%rld = 0.55_8       ! Load ratio
        spec%rite = 2.260_8     ! Ite


!   ----- ROOT BUS, CREW>=11persons, "BRx" ---------------------
    case ("BR1")                                                ! BR1 6t< GVW <= 8t
        spec%w0 = 5186.0_8
        spec%wld = 0.0_8
        spec%crew = 39
        spec%bh = 2.880_8
        spec%bw = 2.072_8
        spec%ric = 0.0_8
        spec%rld = 0.35_8       ! Load ratio
        spec%rite = 0.703_8     ! Ite


    case ("BR2")                                                ! BR2 8t< GVW <= 10t
        spec%w0 = 7837.0_8
        spec%wld = 0.0_8
        spec%crew = 28
        spec%bh = 2.990_8
        spec%bw = 2.315_8
        spec%ric = 0.0_8
        spec%rld = 0.35_8       ! Load ratio
        spec%rite = 1.101_8     ! Ite


    case ("BR3")                                                ! BR3 10t< GVW <= 12t
        spec%w0 = 7901.0_8
        spec%wld = 0.0_8
        spec%crew = 59
        spec%bh = 2.989_8
        spec%bw = 2.312_8
        spec%ric = 0.0_8
        spec%rld = 0.35_8       ! Load ratio
        spec%rite = 1.101_8     ! Ite


    case ("BR4")                                                ! BR4 12t< GVW <= 14t
        spec%w0 = 8654.0_8
        spec%wld = 0.0_8
        spec%crew = 77
        spec%bh = 2.969     
        spec%bw = 2.385_8
        spec%ric = 0.0_8
        spec%rld = 0.35_8       ! Load ratio
        spec%rite = 1.101_8     ! Ite


    case ("BR5")                                                ! BR5 14t< GVW
        spec%w0 = 10203.0_8
        spec%wld = 0.0_8
        spec%crew = 79
        spec%bh = 3.022_8
        spec%bw = 2.490_8
        spec%ric = 0.0_8
        spec%rld = 0.35_8       ! Load ratio
        spec%rite = 1.101_8     ! Ite


!   ----- GENERAL BUS, CREW>=11persons, "Bx" -------------------
    case ("B1")                                                 ! B1 3.5t< GVW <= 6t
        spec%w0 = 3681.0_8
        spec%wld = 0.0_8
        spec%crew = 29
        spec%bh = 2.581_8
        spec%bw = 2.029_8
        spec%ric = 0.15_8
        spec%rld = 0.60_8       ! Load ratio
        spec%rite = 0.315_8     ! Ite


    case ("B2")                                                 ! B2 6t< GVW <= 8t
        spec%w0 = 5622.0_8
        spec%wld = 0.0_8
        spec%crew = 29
        spec%bh = 3.019_8
        spec%bw = 2.197_8
        spec%ric = 0.15_8
        spec%rld = 0.60_8       ! Load ratio
        spec%rite = 0.703_8     ! Ite


    case ("B3")                                                 ! B3 8t< GVW <= 10t
        spec%w0 = 6608.0_8
        spec%wld = 0.0_8
        spec%crew = 49
        spec%bh = 3.105_8
        spec%bw = 2.314_8
        spec%ric = 0.15_8
        spec%rld = 0.60_8       ! Load ratio
        spec%rite = 1.101_8     ! Ite


    case ("B4")                                                 ! B4 10t< GVW <= 12t
        spec%w0 = 8181.0_8
        spec%wld = 0.0_8
        spec%crew = 40
        spec%bh = 3.213_8
        spec%bw = 2.400_8
        spec%ric = 0.45_8
        spec%rld = 0.65_8       ! Load ratio
        spec%rite = 1.101_8     ! Ite


    case ("B5")                                                 ! B5 12t< GVW <= 14t
        spec%w0 = 10198.0_8
        spec%wld = 0.0_8
        spec%crew = 60
        spec%bh = 3.228_8
        spec%bw = 2.490_8
        spec%ric = 0.45_8
        spec%rld = 0.65_8       ! Load ratio
        spec%rite = 1.101_8     ! Ite


    case ("B6")                ! B6 14t< GVW <= 16t
        spec%w0 = 12296.0_8
        spec%wld = 0.0_8
        spec%crew = 57
        spec%bh = 3.449_8
        spec%bw = 2.49_8
        spec%ric = 0.55_8
        spec%rld = 0.65_8       ! Load ratio
        spec%rite = 1.101_8     ! Ite


    case ("B7")                 ! B7 16t< GVW
        spec%w0 = 12757.0_8
        spec%wld = 0.0_8
        spec%crew = 61
        spec%bh = 3.489_8
        spec%bw = 2.490_8
        spec%ric = 0.55_8
        spec%rld = 0.65_8       ! Load ratio
        spec%rite = 1.650_8     ! Ite

    case default
        write (0,'(3A)') " Category Error : [ ", trim(categoryname), " ] is not defined."
        stop
   end select

end subroutine
