
module mshiftp

    interface

        recursive subroutine runmode ( i,vpat,vp,sp,sht,nr,spec,eng,tm,tqc,evi,eti,spi,vi,verr )
            use hvtransfc
            type ( sspec ) :: spec
            type (sengine) :: eng
            type ( stransmission ) :: tm
            type ( stqcurve ) :: tqc
            type ( scycle ) :: vpat
            real(8) vi, evi, eti, vp, verr
            integer nr, i, sp, spi, sht
        end subroutine

        subroutine sftdwn ( i,vpat,vp,sp,spec,eng,tm,tqc,cnt,spout,verr )
            use hvtransfc
            type ( sspec ) :: spec
            type (sengine) :: eng
            type ( stransmission ) :: tm
            type ( stqcurve ) :: tqc
            type ( scycle ) :: vpat
            integer i, sp, spout, cnt
            real(8) vp, verr
        end subroutine

        subroutine sftuptm ( reacc,i,vpat,vp,sp,spec,eng,tm,tqc,cnt,spout,verr )
            use hvtransfc
            type ( sspec ) :: spec
            type (sengine) :: eng
            type ( stransmission ) :: tm
            type ( stqcurve ) :: tqc
            type ( scycle ) :: vpat
            integer i, sp, spout, cnt, reacc
            real(8) vp, verr
        end subroutine

        subroutine sftup ( reacc,i,vpat,vp,sp,spec,eng,tm,tqc,spout,mkt )
            use hvtransfc
            type ( sspec ) :: spec
            type (sengine) :: eng
            type ( stransmission ) :: tm
            type ( stqcurve ) :: tqc
            type ( scycle ) :: vpat
            integer i, sp, spout, reacc, mkt
            real(8) vp
        end subroutine

        integer function startgear ( i, vpat, spec, eng, tm, tqc, stime )
            use hvtransfc
            type ( sspec ) :: spec
            type (sengine) :: eng
            type ( stransmission ) :: tm
            type ( stqcurve ) :: tqc
            type ( scycle ) :: vpat
            integer i, stime
        end function startgear

        subroutine engstat ( ev,et,sp,spec,eng,tm,tqc,vdest,grade,vp,vv,maxt,sw )
            use hvtransfc
            type ( sspec ) :: spec
            type (sengine) :: eng
            type ( stransmission ) :: tm
            type ( stqcurve ) :: tqc
            real(8) ev, et, vdest, maxt, vp, vv, grade
            integer sp, sw
        end subroutine engstat

        function drvfrc ( spec, tm, sp, v1, v2, grade )
            use hvtransfc
            type ( sspec ) :: spec
            type ( stransmission ) :: tm
            integer sp
            real(8) drvfrc, v1, v2, grade
        end function drvfrc

        real(8) function maxtqlin( tqc, rev )
            use hvtransfc
            type (stqcurve) :: tqc
            real(8) rev
        end function maxtqlin

    end interface

end module mshiftp




! *************************************************************
! SUBROUTINE RUNMODE
!  calculation of gear position & vehicle running conditions
!  i : index
!  vpat : test cycle
!  vp : past speed (km/h)
!  sp : past gear position
!  theld : gear holding time (s)
!  nrecur : recursion counter
!  spec : vehicle spec
!  eng : engine data
!  tm : transmission data
!  tqc : maximum torque curve
!  evi : output engine speed
!  eti : output engine torque
!  spi : output gear position
!  vi  : output vehicle speed (km/h)
!  verr : cumulative velocity error
! *************************************************************
recursive subroutine runmode ( i, vpat, vp, sp, theld, nrecur, spec, eng, tm, tqc,     &
                               evi, eti, spi, vi, verr )
    use modconst
    use hvtransfc
    use mshiftp, only : sftup, sftuptm, sftdwn, startgear
    implicit none

    type (scycle) :: vpat
    type (sspec) :: spec
    type (sengine) :: eng
    type (stransmission) :: tm
    type (stqcurve) :: tqc
    integer i, sp, theld, nrecur, spi

    real(8) vp, evi, eti, vi, verr

    real(8) errmin                                              ! minimum error
    integer sw                                                  ! engine condition return code
    integer stime                                               ! starting time length
    real(8) maxt                                                ! maximum engine torque
    real(8) dv                                                  ! velocity difference
    integer reacc                                               ! reacceleration (0:OFF, 1:ON)
    integer spi2, theld2, ct, sw2, s2                           ! for recursion
    real(8) ev2, et2, vi2, verr2
    integer mkt                                                 ! estimated minimum keep time


!   ===== DETECT END OF RECURSIVE CALCULATION ==================
    ct = nrecur - 1
    if( ct < 0 .or. i > vpat%ndat ) then
        spi = sp
        verr = verr
        return
    end if

!   ===== VDEST(I) = 0, VEHICLE STOPS ==========================
    if ( vpat%v(i) <= 0.0_8 ) then                              ! vehicle stops
        vi = vpat%v(i)
        evi = eng%nidle
        eti = 0.0_8
        spi = 0
        verr = verr                                             &
             + ( vpat%v(i-1) + vpat%v(i) ) / 2.0_8              &
             - ( vi + vp ) / 2.0_8
        return
    end if

!   ===== VDEST(I) > 0 =========================================
    mkt = THOLD
    errmin = 0.0_8
    reacc = 0                                                   ! reacceleration
    dv = vpat%v(i) - vp
!   ===== PAST SHIFT POSITION WAS NEUTRAL ======================
    if ( sp == 0 ) then
!       ===== VEHICLE STARTS TO MOVE, DETERMINE GEAR POSITION ==
        if ( vp == 0.0_8 ) then
            spi = startgear ( i, vpat, spec, eng, tm, tqc, stime )
            theld = THOLD - stime
            call engstat ( evi, eti, spi, spec, eng, tm, tqc,   &
                           vpat%v(i), vpat%grade(i), vp, vi, maxt, sw2 )
!       ===== REACCELERATION MODE ==============================
        else if ( dv >= 0 ) then
            reacc = 1
            theld = THOLD
            spi = tm%grb                                        ! shift to bottom gear
!       ===== DECELERATE WITH NEUTRAL POSITION =================
        else
            vi = vpat%v(i)
            evi = eng%nidle
            eti = 0.0_8
            theld = THOLD
            verr = verr                                         &
                 + ( vpat%v(i-1) + vpat%v(i) ) / 2.0_8          &
                 - ( vi + vp ) / 2.0_8
            return
        end if

!   ===== PAST SHIFT POSITION WAS NOT NEUTRAL ==================
    else
!       ----- check engine speed -------------------------------
        ev2 = vpat%v(i) * 60.0_8 * tm%gri(sp)%gr                &
              * tm%fgr / ( 3.6_8 * spec%tcl )

        if ( ( dv >= 0.0_8 ) .and.                          &   ! acc>=0
             ( ev2 < eng%nel(sp) ) .and.                    &   ! ev2 < lower limit
             ( sp > tm%grs ) ) then                             ! sp > start gear
            spi = tm%grb                                        ! shift to bottom gear
            reacc = 1                                           ! reacceleration mode ON
        else
            spi = sp                                            ! normal condition
        end if
    end if

!   ===== CALCULATE ENGINE RUNNING CONDITION ===================
    call engstat ( evi, eti, spi, spec, eng, tm, tqc,           &
                   vpat%v(i), vpat%grade(i), vp, vi, maxt, sw )

!   ===== CASE THAT CANNOT CHANGE GEAR, DECELERATION  ==========
    if ( ( theld < THOLD ) .or. ( dv < 0.0_8 ) ) then

        if ( sw == 4 .or. sw == 3 ) then                        ! clutch off(SW=4) or lower limit(SW=3)
            spi = 0                                             ! neutral
            theld = THOLD                                       ! enable gear change
            return
        end if

        if ( sw == 2 .and. spi < tm%ngr ) then                  ! overrun(SW=2)
            vi = 0.0_8                                          ! cannot continue calculation
    
            return                                              ! (RETURN ABNORMAL RESULT)
        end if

        verr2 = verr                                            &
                + ( vpat%v(i-1) + vpat%v(i) ) / 2.0_8           &
                - ( vi + vp ) / 2.0_8
        theld2 = theld + 1
        call runmode ( i+1, vpat, vi, spi, theld2, ct, spec,    &
                       eng, tm, tqc, ev2, et2, spi2, vi2, verr2 )
        verr = verr2
        theld = theld + 1                                       ! increment gear hold time

        return

    end if

!   ===== CASE THAT CAN CHANGE GEAR POSITION ===================
!   ===== REACCELERATION =======================================
    if ( reacc == 1 ) then
!       ----- TRY SHIFT-UP (WITH TORQUE MARGIN) ----------------
        call sftuptm ( reacc, i, vpat, vp, spi, spec,           &
                       eng, tm, tqc, ct, spi2, verr2 )
        if ( spi < spi2 ) then
            spi = spi2
!       ----- FIND OPTIMAL GEAR THAT CAN KEEP 3 SECONDS --------
        else
            call sftup( reacc, i, vpat, vp, spi, spec,          &
                        eng, tm, tqc, s2, mkt )
            spi = s2
        end if

!   ===== OVER RUN. SHIFT-UP REQUIRED ==========================
    else if ( sw == 2 ) then
        if ( spi < tm%ngr ) then
!           ----- TRY SHIFT-UP (WITH TORQUE MARGIN) ------------
            call sftuptm ( reacc, i, vpat, vp, spi, spec,       &
                           eng, tm, tqc, ct, s2, verr2 )
!           ----- FIND OPTIMAL GEAR THAT CAN KEEP 3 SECONDS ----
            if ( spi == s2 ) then
                call sftup( reacc, i, vpat,  vp, spi, spec,     &
                            eng, tm, tqc, s2, mkt )
            end if
!           ----- ENGINE RUNNING CONDITION ---------------------
            call engstat ( evi, eti, s2, spec, eng, tm, tqc,    &
                           vpat%v(i), vpat%grade(i), vp, vi, maxt, sw )
            spi = s2
            theld2 = 1 + ( THOLD - mkt )
        else
            theld2 = theld + 1                                  ! top gear
        end if
!       ----- CALCULATE CUMLATIVE ERROR ------------------------
        verr2 = verr                                            &
                + ( vpat%v(i-1) + vpat%v(i) )                   &
                / 2.0_8 - ( vi + vp ) / 2.0_8
        call runmode ( i+1, vpat, vi, spi, theld2, ct, spec,    &
                       eng, tm, tqc, ev2, et2, spi2, vi2, verr2 )
        errmin = verr2

!   ===== ALL OTHER CASES ======================================
    else
!       ----- KEEPS SAME GEAR POSITION -------------------------
        theld2 = theld + 1
        verr2 = verr                                            &
                + ( vpat%v(i-1) + vpat%v(i) )                   &
                / 2.0_8 - ( vi + vp ) / 2.0_8

        call runmode ( i+1, vpat, vi, spi, theld2, ct, spec,    &
                       eng, tm, tqc, ev2, et2, spi2, vi2, verr2 )
        errmin = verr2

!       ----- TRY SHIFT-UP (WITH TORQUE MARGIN) ----------------
        call sftuptm ( reacc, i, vpat, vp, spi, spec, eng,      &
                       tm, tqc, ct, spi2, verr2 )
        if ( spi2 > spi .and. errmin >= verr+verr2 ) then
            spi = spi2
            errmin = verr+verr2
        end if

!       ----- TRY SHIFT-DOWN -----------------------------------
!        if ( sp == spi .and. sw == 1 ) then   ! 2022.10.05_̃IWi
        if ( sp == spi .and. (sw == 1 .or. sw == 3) ) then   ! 2022.10.05 ŒENG]ȉ̏ꍇisw==3jɂVtg_E݂dlɕύX
            call sftdwn ( i, vpat, vp, spi, spec, eng,          &
                          tm, tqc,  ct, s2, verr2 )
            if ( spi > s2 .and. errmin >= verr+verr2 ) then
                spi = s2
                errmin = verr+verr2
            end if
        end if
    end if

!   ===== FINAL RUNNING CONDITION ==============================
    call engstat ( evi, eti, spi, spec, eng, tm, tqc,           &
                   vpat%v(i), vpat%grade(i), vp, vi, maxt, sw )

    if ( dv >= 0.0_8 .and. evi < eng%nes ) then                ! reacc by bottom gear
        evi = eng%nes                                          ! keep starting speed
    else if ( evi < eng%nec ) then                             ! clutch off
        evi = eng%nidle
        eti = 0.0_8
        spi = 0
    end if

    if ( spi == 0 ) then                                        ! neutral
        theld = THOLD                                           ! enable gear change
    else if ( sp /= spi ) then
        theld = 1 + ( THOLD - mkt )                             ! reset hold time
    else
        theld = theld + 1                                       ! increment hold time
    end if
    verr = errmin

end subroutine runmode



! *************************************************************
! FUNCTION STARTGEAR
!  calculation of starting gear
!  i : index
!  vpat : test cycle
!  spec : vehicle spec
!  eng : engine data
!  tm : transmission data
!  tqc : maximum torque curve
!  stime : time length of starting
!  startgear : starting gear
! *************************************************************
integer function startgear ( i, vpat, spec, eng, tm, tqc, stime )
    use hvtransfc
    use mshiftp, only : engstat
    implicit none

    type (sspec) :: spec
    type (sengine) :: eng
    type (stransmission) :: tm
    type (stqcurve) :: tqc
    type (scycle) :: vpat
    integer i, stime
    integer i2, flg, flg2, sp, sw
    real(8) ev, et, vi, maxt

    flg2 = 0
    do sp = tm%grs, 1, -1                                       ! test starting gear
        i2 = i
        flg = 0
        stime = 0
        startgear = sp
        do while ( flg == 0 )
            stime = stime + 1                                   ! predict running condition
            call engstat ( ev, et, sp, spec, eng, tm, tqc, vpat%v(i2),  &
                           vpat%grade(i2), vpat%v(i2-1), vi, maxt, sw )

            if ( sw == 0 .or. sw == 2 ) then
                flg = 1                                         ! normal(0) or overrun(2)
                flg2 = 1                                        ! fix startgear
            else if ( sw == 1 .or. sw == 3 ) then
                flg = 1                                         ! lack of torque(1), out of range(3)
            else
                i2 = i2 + 1                                     ! continue
            end if
        end do
        if ( flg2 == 1 ) exit                                   ! exit loop
    end do

end function startgear



! *************************************************************
! SUBROUTINE SFTUPTM
!  calculation of shift-up (with torque margin)
!  reacc : reacceleration mode ( 1: reacceleration, 0: non )
!  i : index
!  vpat : test cycle
!  vp : past speed
!  sp : past gear position
!  spec : vehicle spec
!  eng : engine data
!  tm : transmission data
!  tqc : maximum torque curve
!  cnt : recursion counter
!  spout : output gear position
!  minerr : cumulative error
! *************************************************************
subroutine sftuptm( reacc, i, vpat, vp, sp, spec, eng, tm, tqc, cnt, spout, minerr )
    use hvtransfc
    use mshiftp, only : engstat
    implicit none

    type (sspec) :: spec
    type (sengine) :: eng
    type (stransmission) :: tm
    type (stqcurve) :: tqc
    type (scycle) :: vpat
    integer reacc, i, sp, cnt, spout
    real(8) vp, minerr
    integer s, sw, x
    real(8) verr, vold, vv, tqmargin
    real(8) ev, et, maxt

    spout = sp
    minerr = 0.0_8

    do s = sp + 1, tm%ngr
        verr = 0.0_8

        if ( ( s > sp + tm%skiplimit ) .and. (reacc == 0) ) then
            exit                                                ! check skip limit
        end if

        call engstat( ev, et, s, spec, eng, tm, tqc,            &
                      vpat%v(i), vpat%grade(i), vp, vv, maxt, sw )

        if ( sw == 2 .or. sw == 3 .or. sw == 5 ) cycle          ! overrun(2), out-of-range(3), starting(5)

        if ( et > 0.0_8 ) then
            tqmargin = maxt / et                                ! torque margin
        else
            tqmargin = 100.0_8                                  ! if (te < 0)
        end if

        if ( tqmargin >= tm%gri(s)%tmargin .and. sw == 0 ) then ! try shift up
            vold = vp
            do x = 0, cnt                                       ! predict running condition
                call engstat ( ev, et, s, spec, eng, tm, tqc, vpat%v(i+x),  &
                               vpat%grade(i+x), vold, vv, maxt, sw )
                if( sw == 2 .or. sw == 3 ) exit                 ! overrun(2), out-of-range(3)
                verr = verr                                     &
                      + ( vpat%v(i+x) + vpat%v(i+x-1) ) / 2.0_8 &
                      - ( vv + vold ) / 2.0_8
                vold = vv
            end do

            if( sw == 2 .or. sw == 3 ) then
                cycle                                           ! overrun(2), out-of-range(3)
            else if ( spout == sp ) then
                spout = s                                       ! update gear position
                minerr = verr                                   ! update error
            else if( minerr >= verr ) then
                spout = s
                minerr = verr
            end if
        end if
    end do

end subroutine sftuptm



! *************************************************************
!  SUBROUTINE SFTUP
!  calculation of shift-up (except torque margin)
!  reacc : reacceleration mode ( 1: reacceleration, 0: non )
!  i : index
!  vpat : test cycle
!  vp : past speed
!  sp : past gear position
!  spec : vehicle spec
!  eng : engine data
!  tm : transmission data
!  tqc : maximum torque curve
!  spout : output gear position
!  mkt : estimated minimum gear hold time
! *************************************************************
subroutine sftup ( reacc, i, vpat, vp, sp, spec, eng, tm, tqc, spout, mkt )
    use hvtransfc
    use mshiftp, only : engstat
    use modconst
    implicit none

    type (sspec) :: spec
    type (sengine) :: eng
    type (stransmission) :: tm
    type (stqcurve) :: tqc
    type (scycle) :: vpat
    integer reacc, i, sp, spout, mkt
    real(8) vp
    integer x, y, z, sw
    real(8) verr, minerr, vold, vv, maxt, ev, et

    spout = sp
    do x = THOLD, 1, -1                                         ! gear hold time
        minerr = 0.0_8
        do y = sp + 1, tm%ngr                                   ! gear position
            vold = vp
            verr = 0.0_8

            do z = 0, x - 1                                     ! predict running condition
                call engstat( ev, et, y, spec, eng, tm, tqc,    &
                              vpat%v(i+z), vpat%grade(i+z),     &
                              vold, vv, maxt, sw )
                if( sw == 2 .or. sw == 3 ) exit                 ! overrun(2), out-of-range(3)
                verr = verr                                     &
                     + ( vpat%v(i+z) + vpat%v(i+z-1) ) / 2.0_8  &
                     - ( vv + vold ) / 2.0_8
                vold = vv
            end do

            if( sw == 2 .or. sw == 3 ) then
                cycle                                           ! overrun(2), out-of-range(3)
            else if ( spout == sp ) then
                spout = y
                minerr = verr
                if ( minerr == 0.0_8 .and. reacc == 0 ) exit
            else if ( minerr > verr ) then
                spout = y
                minerr = verr
            end if
        end do
        if ( spout /= sp ) then
            mkt = x
            exit
        end if
    end do

end subroutine sftup



! *************************************************************
! SUBROUTINE SFTDWN
!  calculation of shift down
!  i : index
!  vpat : test cycle
!  vp : past speed
!  sp : past gear position
!  spec : vehicle spec
!  eng : engine data
!  tm : transmission data
!  tqc : maximum torque curve
!  cnt : recursion counter
!  spout : output gear position
!  minerr : cumulative error
! *************************************************************
subroutine sftdwn ( i, vpat, vp, sp, spec, eng, tm, tqc, cnt, spout, minerr )
    use hvtransfc
    use mshiftp, only : engstat
    implicit none

    type (sspec) :: spec
    type (sengine) :: eng
    type (stransmission) :: tm
    type (stqcurve) :: tqc
    type (scycle) :: vpat
    integer i, sp, cnt, spout
    real(8) vp, minerr
    integer s, sw, x
    real(8) verr, vold, vnew, ev, et, maxt

    spout = sp                                                  ! initial position
    do s = tm%grb, sp - 1
        vold = vp
        verr = 0.0_8
        do x = 0, cnt
            call engstat ( ev, et, s, spec, eng, tm, tqc, vpat%v(i+x),  &
                           vpat%grade(i+x), vold, vnew, maxt, sw )

            if( sw == 2 .or. sw == 3 ) exit                     ! overrun(2), out-of-range(3)
            verr = verr                                         &
                 + ( vpat%v(i+x) + vpat%v(i+x-1) ) / 2.0_8      &
                 - ( vnew + vold ) / 2.0_8
            vold = vnew
        end do

        if( sw == 2 .or. sw == 3 ) then
            cycle                                               ! overrun(2), out-of-range(3)
        else if ( spout == sp .or. minerr >= verr ) then
            spout = s                                           ! update gear position
            minerr = verr                                       ! minimum cumulative error
        end if
    end do

end subroutine sftdwn



! *************************************************************
! FUNCTION DRVFRC
!  calculate driving force
!  spec : vehicle spec
!  mis  : transmission
!  sp   : gear position
!  v1   : current speed (km/h)
!  v2   : target speed (km/h)
!  grade  : grade (%)
! *************************************************************
real(8) function drvfrc ( spec, tm, sp, v1, v2, grade )
    use hvtransfc
    implicit none

    type ( sspec ) :: spec
    type ( stransmission ) :: tm
    integer sp
    real(8) v1, v2, grade
    real(8) rr, mass, acc, fgrade

    rr =  spec%mur * spec%wt + spec%mua * ( v2 * v2 )
    mass = spec%wt + tm%gri(sp)%dw
    acc = ( v2 - v1 ) / ( 3.6_8 * 1.0_8 )
    fgrade = spec%wt * dsin( datan( grade / 100.0_8 ) ) * 9.8_8

    drvfrc = rr + mass * acc  + fgrade                          ! driving force (N)
    
 !   write(7777,*) 'v2, rr = ', v2, rr , ' [km/h]  [N]'

end function drvfrc



! *************************************************************
! FUNCTION MAXTQLIN
!  liniear interpolation of torque curve
!  tqc      : torque curve
!  rev      : input engine speed, rpm
!  maxtqlin : interpolated torque, (Nm)
! *************************************************************
real(8) function maxtqlin( tqc, rev )
    use hvtransfc
    implicit none

    type ( stqcurve ) :: tqc                                    ! torque curve array
    real(8) rev                                                 ! input engine speed, rpm
    integer x                                                   ! index

    if ( rev < tqc%rev(1) ) then
        x = 1                                                   ! rev<r[1]
    else if ( rev >= tqc%rev(tqc%ndata) ) then
        x = tqc%ndata -1                                        ! rev>=r[ndata]
    else
        do x = 1, tqc%ndata
            if ( rev >= tqc%rev(x) .and. rev < tqc%rev(x+1) ) then
                exit
            end if
        end do
    end if

    maxtqlin = tqc%tq(x) + ( tqc%tq(x+1)-tqc%tq(x) )            &
             * ( rev-tqc%rev(x) ) / ( tqc%rev(x+1)-tqc%rev(x) )

end function maxtqlin



! *************************************************************
! SUBROUTINE ENGSTAT
!  calculate engine running condition
!  output engine speed, engine torque, maximum torque,
!  vehicle speed, engine condition
!  ev : engine speed
!  et : engine torque
!  sp : gear position
!  spec : vehicle spec
!  eng : engine data
!  tm : transmission data
!  tqc : maximum torque curve
!  vdest : target speed (km/h)
!  grade : grade (%)
!  vp : past speed (km/h)
!  vv : output vehicle speed
!  maxt : maximum torque
!  sw : return code
!   sw=0 engine speed is in range
!   sw=1 lack of torque
!   sw=2 engine overrun
!   sw=3 engine speed under limit
!   sw=4 clutch off
!   sw=5 starting,  Ne <= Nes, shift = 1st or 2nd
! *************************************************************
subroutine engstat ( ev, et, sp, spec, eng, tm, tqc, vdest, grade, vp, vv, maxt, sw )
    use hvtransfc
    use mshiftp, only : maxtqlin, drvfrc
    implicit none

    type ( sspec ) :: spec
    type (stransmission) :: tm
    type (sengine) :: eng
    type ( stqcurve ) :: tqc
    real(8) vp, vdest, vv, grade
    real(8) ev, et, maxt
    integer sp, sw

    real(8) dv, tqdif, df, ds, tgr, tegr
    integer flg, iter    ! 200110 add.

!   ===== CALCULATE INITIAL CONDITION ===================================================
    vv = vdest                                                  ! initialize output speed
    df = drvfrc ( spec, tm, sp, vp, vv, grade )                 ! required driving force
    tgr = tm%gri(sp)%gr * tm%fgr                                ! gear ratio
    tegr = tm%gri(sp)%egr * tm%efgr                             ! efficiency
    if ( df >= 0.0_8 ) then
        et = df * spec%rt / ( tgr * tegr )                      ! engine torque (when driving)
    else
        et = df * ( spec%rt  * tegr ) / tgr                     ! engine torque (when braking)
    end if

    ev = ( vv  * 60.0_8 ) * ( tgr ) / ( spec%tcl * 3.6_8 )      ! engine speed
    dv = vv - vp                                                ! difference of speed

!   -----------------------------------------------------------
!   decelerate
!   -----------------------------------------------------------
    if ( dv < 0.0_8 ) then
        if ( ev < eng%nec ) then                                ! clutch off
            ev = eng%nidle                                      ! idling speed
            et = 0.0_8
            sw = 4                                              ! [ RETURN CODE = 4 ] clutch off
        else
            sw = 0                                              ! [ RETURN CODE = 0 ] normal cond.
        end if
        maxt = maxtqlin ( tqc, ev )
        return
    end if

!   -----------------------------------------------------------
!   accelerate
!   -----------------------------------------------------------
    if ( ev < eng%nel(sp) ) then                                ! engine speed is under lower limit  
        if ( sp > tm%grs ) then                                 ! except start gear
!            vv = 0.0_8                                          ! stop running
            maxt = maxtqlin ( tqc, eng%nel(sp) )
            sw = 3                                              ! [ RETURN CODE = 3 ] no selectable gear
            return
        else                                                    ! start gear
            ev = eng%nes                                        ! starting engine speed
            maxt = maxtqlin ( tqc, ev )
            if ( et > maxt ) then
                sw = 1                                          ! [ RETURN CODE = 1 ] lack of torque
            else
                sw = 5                                          ! [ RETURN CODE = 5 ] starting
                return
            end if
        end if

    else if ( ev >= eng%nex ) then                              ! engine speed is over upper limit
        if ( sp == tm%ngr )then                                 ! highest gear
            ev = eng%nex                                        ! engine speed limit
        end if
        vv = ev * spec%tcl * 3.6_8 / ( 60.0_8 * tgr )           ! vehicle speed
        maxt = maxtqlin ( tqc, ev )
        df = drvfrc ( spec, tm, sp, vp, vv, grade )

        if ( df >= 0.0_8 ) then
            et = df * spec%rt / ( tgr * tegr )                  ! engine torque(when driving)
        else
            et = df * ( spec%rt  * tegr ) / tgr                 ! engine torque(when braking)
        end if

        if ( (sp == tm%ngr) .and. ( et > maxt ) .and. ( maxt > 0.0_8 ) ) then
            sw = 1                                              ! [ RETURN CODE = 1 ], lack of torque
        else
            sw = 2                                              ! [ RETURN CODE = 2 ], engine overrun
            return
        end if

    else                                                        ! engine speed is in range
        maxt = maxtqlin ( tqc, ev )                             ! MINIMUM <= ENGINE SPEED < MAXIMUM
        if ( et > maxt .and. maxt > 0.0_8 ) then
            sw = 1                                              ! [ RETURN CODE = 1 ], lack of torque
        else
            sw = 0                                              ! [ RETURN CODE = 0 ], normal cond.
            return
        end if
    end if

!   -----------------------------------------------------------
!   convergence calculation of vehicle speed
!   -----------------------------------------------------------
    ds = 1.0_8
    flg = 0
!    do while ( flg == 0 )
    do iter = 1, 9999
        if(flg /= 0)exit
        df = drvfrc ( spec, tm, sp, vp, vv, grade )

        if ( df >= 0.0_8 ) then
            et = df * spec%rt / ( tgr * tegr )                  ! engine torque(when driving)
        else
            et = df * ( spec%rt  * tegr ) / tgr                 ! engine torque(when braking)
        end if

        ev = vv * 60.0_8 * tgr / ( spec%tcl * 3.6_8 )           ! engine speed

        if ( sp > tm%grb .and. ev < eng%nel(sp) ) then
            maxt = maxtqlin ( tqc, eng%nel(sp) )                ! engine speed is under lower limit
        else if ( ev < eng%nes ) then
            ev = eng%nes                                        ! starting engine speed
            maxt = maxtqlin ( tqc, ev ) 
            if(maxt>et)return                                   ! 220907 add.
        else
            maxt = maxtqlin ( tqc, ev )
        end if

        tqdif = maxt - et
        if ( ( tqdif < 1.0E-6 ) .and. ( tqdif >= 0.0_8 ) ) then ! check convergence calculation result
            flg = 1
        else if ( tqdif < 0.0_8 ) then
            vv = vv - ds                                        ! decrease target speed
        else
            ds = ds / 2.0_8                                     ! increase target speed
            vv = vv + ds
        end if
        if(iter>9998)write(*,*)'Warning : convergence cal. of vehicle speed'
    end do

!   -----------------------------------------------------------
!   final engine running condition
!   -----------------------------------------------------------
    if ( ev < eng%nel(sp) ) then                                ! engine speed is under the lower limit
        vv = 0.0_8                                              ! cannot continue calculation
        maxt = maxtqlin ( tqc, eng%nel(sp) )
        sw = 3                                                  ! [ RETURN CODE = 3 ], out of range
    else if ( ev >= eng%nex ) then                              ! engine overrun
        vv = 0.0_8                                              ! cannot continue calculation
        maxt = maxtqlin ( tqc, eng%nex )
        sw = 2                                                  ! [ return code = 2 ], engine overrun
    end if

end subroutine engstat
